/* -*- mode:c; mode:fold -*- */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <tiffio.h>

#undef debug
/* #define debug */

#undef write_bar
/* #define write_bar */


/* Bitmap Header */ /*{{{*/
typedef unsigned char   BYTE;
typedef unsigned short  WORD;
typedef         uint32  DWORD;
typedef          int32  LONG;

#pragma pack(1)
typedef struct {
  WORD    bfType;
  DWORD   bfSize;
  WORD    bfReserved1;
  WORD    bfReserved2;
  DWORD   bfOffBits;
} BitmapFileHeader;

typedef struct {
  DWORD      biSize;
  LONG       biWidth;
  LONG       biHeight;
  WORD       biPlanes;
  WORD       biBitCount;
  DWORD      biCompression;
  DWORD      biSizeImage;
  LONG       biXPelsPerMeter;
  LONG       biYPelsPerMeter;
  DWORD      biClrUsed;
  DWORD      biClrImportant;
} BitmapInfoHeader;
/*}}}*/

/**********************************************/
 void writeBitmapHeaders(FILE *F, int w, int h)
/**********************************************/ /*{{{*/
{
  BitmapFileHeader bmfh = { 'B'+'M'*0x100, -1, 0, 0, sizeof(BitmapFileHeader)+sizeof(BitmapInfoHeader) };
  BitmapInfoHeader bmih = { sizeof(BitmapInfoHeader), -1, -1, 1, 24, 0, -1, 2835, 2835, 0, 0 };
  bmfh.bfSize = sizeof(BitmapFileHeader) + sizeof(BitmapInfoHeader) + w*h*3;
  bmih.biWidth = w;
  bmih.biHeight = h;
  bmih.biSizeImage = w*h*3;
  fwrite(&bmfh, sizeof(BitmapFileHeader), 1, F);
  fwrite(&bmih, sizeof(BitmapInfoHeader), 1, F);
}
/*}}}*/

/*************/
 double arcsin(double x)
/*************/ /*{{{*/
{
  return x>=0
    ?  atan( x/sqrt(1-x*x))
    : -atan(-x/sqrt(1-x*x));
}
/*}}}*/

typedef struct {
  char* tifFilename;
  char* outputFilename;
  int size;
  double d;
  double q;
  double phi0;
  double lambda0;
  double phiSun;
  double tSun;
  double shadowFactor;
  double brightnessFactor;
} GlobProjParams;

/**********************/
 void initDefaultParams(GlobProjParams *par)
/**********************/ /*{{{*/
{
  par->tifFilename = malloc(256);
  par->outputFilename = malloc(256);
  strcpy(par->tifFilename, "land_ocean_ice_cloud_8192.tif");
  strcpy(par->outputFilename, "Erde.bmp");
  par->size =
#ifndef write_bar
    4800 /*pix*/;
#else
    400 /*pix*/;
#endif
  par->d = 2 /*R[Earth]*/;
  par->q = 2.5 /*R[Earth]*/;
  par->phi0 = 48. /*deg*/;
  par->lambda0 = 12. /*deg*/;
  par->phiSun = 23.5 /*deg*/;
  par->tSun = 12. /*h*/;
  par->shadowFactor = 0.1;
  par->brightnessFactor = 1.2;
}
/*}}}*/

/* GlobProj */
GlobProjParams par;
double d2;
double cos_phi0, sin_phi0, cos_lambda0, sin_lambda0;
double cos_phiSun, sin_phiSun, cos_lambdaSun, sin_lambdaSun;
/* TIFF */
uint32 w, h, *raster;

/******************/
 int getGlobProjRGB(double xi, double eta, BYTE *R, BYTE *G, BYTE *B)
/******************/ /*{{{*/
{
  double rho2, D, t, x, y, z, X, phi, lambda;
  uint32 xpix, ypix, r00, r01, r10, r11;
  double q00, q01, q10, q11, r, g, b;
  double cos_alphaSun;

  *R = 0;
  *G = 0;
  *B = 0;

  /* (xi, eta): coordinates on projection plane */
  rho2 = xi*xi + eta*eta;
  D = d2 - d2*rho2 + rho2;
  if(D < 0)
    return 0;
  t = (rho2+sqrt(D)) / (d2+rho2);
  x = t * par.d;
  y = xi - t*xi;
  z = eta - t*eta;

  /* rotation */
  X = cos_phi0*x    - sin_phi0*z;
  z = sin_phi0*x    + cos_phi0*z;
  x = cos_lambda0*X - sin_lambda0*y;
  y = sin_lambda0*X + cos_lambda0*y;

  /* coordinates on earth */
  phi = arcsin(z);
  lambda = atan2(y, x);
  if(lambda >= M_PI)
    lambda -= 2*M_PI;

  x = w/2*(1+lambda/M_PI);
  y = h/2*(1+ phi*2/M_PI);
  xpix = (int)(x);
  ypix = (int)(y);

  if(xpix < w-1  &&  ypix < h-1)
  {
    r00 = raster[ ypix   *w + xpix  ];
    r01 = raster[ ypix   *w + xpix+1];
    r10 = raster[(ypix+1)*w + xpix  ];
    r11 = raster[(ypix+1)*w + xpix+1];

    q00 = (xpix+1 - x) * (ypix+1 - y);
    q01 = (x - xpix)   * (ypix+1 - y);
    q10 = (xpix+1 - x) * (y - ypix);
    q11 = (x -xpix)    * (y - ypix);

    r =   q00*(BYTE)( (r00 & 0x0000FF))
        + q01*(BYTE)( (r01 & 0x0000FF))
        + q10*(BYTE)( (r10 & 0x0000FF))
        + q11*(BYTE)( (r11 & 0x0000FF));
    g =   q00*((r00 & 0x00FF00) >>  8)
        + q01*((r01 & 0x00FF00) >>  8)
        + q10*((r10 & 0x00FF00) >>  8)
        + q11*((r11 & 0x00FF00) >> 8);
    b =   q00*((r00 & 0xFF0000) >> 16)
        + q01*((r01 & 0xFF0000) >> 16)
        + q10*((r10 & 0xFF0000) >> 16)
        + q11*((r11 & 0xFF0000) >> 16);
  }
  else
  {
    if(xpix < w  &&  ypix < h)
    {
       r00 = raster[ypix*w + xpix];
       r = (BYTE)((r00 & 0x0000FF)      );
       g =       ((r00 & 0x00FF00) >>  8);
       b =       ((r00 & 0xFF0000) >> 16);
    }
    else
      return 0;
  }

  /* shadow */
  cos_alphaSun =   cos(phi)*cos_phiSun * (cos(lambda)*cos_lambdaSun + sin(lambda)*sin_lambdaSun)
                 + sin(phi)*sin_phiSun;
  if(cos_alphaSun < 0)
  {
    r *= par.shadowFactor;
    g *= par.shadowFactor;
    b *= par.shadowFactor;
  }
  else
  {
    r *= par.shadowFactor + (cos_alphaSun*(par.brightnessFactor - par.shadowFactor));
    g *= par.shadowFactor + (cos_alphaSun*(par.brightnessFactor - par.shadowFactor));
    b *= par.shadowFactor + (cos_alphaSun*(par.brightnessFactor - par.shadowFactor));
  }
#ifdef debug
  fprintf(stderr, "%.3f %.3f  %.4f  %.2f %.2f  %.4f\n", xi, eta, y/cos(phi), lambda*180.0/M_PI, phi*180.0/M_PI, cos_alphaSun);
#endif

  if(r > 255)
    r = 255;
  if(g > 255)
    g = 255;
  if(b > 255)
    b = 255;
  *R = (BYTE)(r);
  *G = (BYTE)(g);
  *B = (BYTE)(b);
  return 1;
}
/*}}}*/

FILE *GlobProjBMPfile;

/*********************/
 void writeGlobProjBMP()
/*********************/ /*{{{*/
{
  double lambdaSun;
  double xi, eta;
  BYTE R, G, B;
  int x, y;
  int WH = par.size;
  double d = par.d;
  double q;

  /* set GlobProj global variables */
  d2 = d*d;
  q = par.q > 0 ? par.q : 2*d / sqrt(d2 - 1);
  cos_phi0 = cos(par.phi0 / 180 * M_PI);
  sin_phi0 = sin(par.phi0 / 180 * M_PI);
  cos_lambda0 = cos(par.lambda0 / 180 * M_PI);
  sin_lambda0 = sin(par.lambda0 / 180 * M_PI);

  lambdaSun = par.lambda0 + 180.0 - 15 * par.tSun;
  cos_phiSun = cos(par.phiSun / 180 * M_PI);
  sin_phiSun = sin(par.phiSun / 180 * M_PI);
  cos_lambdaSun = cos(lambdaSun / 180 * M_PI);
  sin_lambdaSun = sin(lambdaSun / 180 * M_PI);

  /* write bitmap */
  for(y=0; y<WH; y++)
  {
    eta = q*((double)(y)/WH-0.5);
    printf("%.1f%%\r", y*100./WH);
    for(x=0; x<WH; x++)
    {
      xi = q*((double)(x)/WH-0.5);
      getGlobProjRGB(xi, eta, &R, &G, &B);
      fwrite(&B, 1, 1, GlobProjBMPfile);
      fwrite(&G, 1, 1, GlobProjBMPfile);
      fwrite(&R, 1, 1, GlobProjBMPfile);
    }
#ifdef debug
    fprintf(stderr, "\n");
#endif
  }
  printf("\n");
}
/*}}}*/

/**************/
 void parseParams(int argc, char** argv, GlobProjParams *par)
/**************/ /*{{{*/
{
  int i;

  initDefaultParams(par);

  for(i=1; i<argc; i++)
  {
    if(strcmp(argv[i], "-h") == 0)
    {
      printf("Usage:\n");
      printf("\t%s [-h] [-i <inputfile>] [-o <outputfile>] [-s <size>] [-d <d>] [-q <q>] [-phi <phi0>] [-lambda <lambda0>] [-phiSun <phiSun>] [-sF <shadowFactor>] [-bF <brightnessFactor>]\n", argv[0]);
      printf("Options:\n");
      printf("\t-h\tshow this help\n");
      printf("\t-i\tuse <inputfile> as TIFF inputfile\n\t\tdefault <inputfile> = %s\n", par->tifFilename);
      printf("\t-o\tuse <outputfile> as BMP outputfile (filename is generated from other parameters if <outputfile> = \"auto\")\n\t\tdefault <outputfile> = %s\n", par->outputFilename);
      printf("\t-s\tcalculate image with size <size> pixels\n\t\tdefault <size> = %d\n", par->size);
      printf("\t-d\tcalculate image from distance <d> Earth radii\n\t\tdefault <d> = %.2f\n", par->d);
      printf("\t-q\tcalculate image with scale <q> Earth radii in projection plane, if greater than 0;\n\t\tif <q> is zero or negative: scale image automatically\n\t\tdefault <q> = %.2f\n", par->q);
      printf("\t-phi\tcalculate image centered above geographic lattitude <phi0> degrees\n\t\tdefault <phi0> = %.2f\n", par->phi0);
      printf("\t-lambda\tcalculate image centered above geographic longitude <lambda0> degrees\n\t\tdefault <lambda0> = %.2f\n", par->lambda0);
      printf("\t-phiSun\tcalculate shadow with sun's geographic lattitude <phiSun> degrees (in reality, |<phiSun>| is less than 23.5)\n\t\tdefault <phiSun> = %.2f\n", par->phiSun);
      printf("\t-tSun\tcalculate shadow at local time <tSun> hours\n\t\tdefault <tSun> = %.2f\n", par->tSun);
      printf("\t-sF\tcalculate shadow with shadow factor <shadowFactor>\n\t\tdefault <shadowFactor> = %.2f\n", par->shadowFactor);
      printf("\t-bF\tcalculate shadow with brightness factor <brightnessFactor>\n\t\tdefault <brightnessFactor> = %.2f\n", par->brightnessFactor);
      exit(0);
    }
  }
  for(i=2; i<argc; i++)
  {
    char *arg = argv[i-1], *value = argv[i];
    if(strcmp(arg, "-i") == 0)
      strcpy(par->tifFilename, value);
    if(strcmp(arg, "-o") == 0)
      strcpy(par->outputFilename, value);
    if(strcmp(arg, "-s") == 0)
      par->size = atoi(value);
    if(strcmp(arg, "-d") == 0)
      par->d = atof(value);
    if(strcmp(arg, "-q") == 0)
      par->q = atof(value);
    if(strcmp(arg, "-phi") == 0)
      par->phi0 = atof(value);
    if(strcmp(arg, "-lambda") == 0)
      par->lambda0 = atof(value);
    if(strcmp(arg, "-phiSun") == 0)
      par->phiSun = atof(value);
    if(strcmp(arg, "-tSun") == 0)
      par->tSun = atof(value);
    if(strcmp(arg, "-sF") == 0)
      par->shadowFactor = atof(value);
    if(strcmp(arg, "-bF") == 0)
      par->brightnessFactor = atof(value);
  }
  if(strcmp(par->outputFilename, "auto") == 0)
    sprintf(par->outputFilename,
#ifndef write_bar
            "GlobProj_%02.0flat_%02.0flon_%02.0flatSun_%04.1fhSun.bmp", par->phi0, par->lambda0, par->phiSun, par->tSun);
#else
            "GlobProj_%02.0flat_%02.0flon_%02.0flatSun.bmp", par->phi0, par->lambda0, par->phiSun);
#endif
}
/*}}}*/

/*********/
 int main(int argc, char** argv)
/*********/ /*{{{*/
{
  TIFF *tif;
  char *filename;
  char cmd[1024];

  parseParams(argc, argv, &par);

  printf("reading %s\n", par.tifFilename);
  tif = TIFFOpen(par.tifFilename, "r");
  if(tif) {
    TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &w);
    TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &h);
    raster = (uint32*)_TIFFmalloc(w*h*sizeof(uint32));
    if(raster != NULL)
    {
      if(TIFFReadRGBAImage(tif, w, h, raster, 0))
      {
        filename = par.outputFilename;
        GlobProjBMPfile = fopen(filename, "wb");
        writeBitmapHeaders(GlobProjBMPfile, par.size, par.size *
#ifndef write_bar
                           1);
#else
                           12);
        for(par.tSun=22.0; par.tSun>-0.01; par.tSun-=2.0)
#endif
          writeGlobProjBMP();

        fclose(GlobProjBMPfile);
        sprintf(cmd, "convert -quality 70 %s %s.jpg", filename, filename);
        printf("%s\n", cmd);
        system(cmd);
        sprintf(cmd, "ls -l %s %s.jpg", filename, filename);
        system(cmd);
      }
      _TIFFfree(raster);
    }
    TIFFClose(tif);
  }

  return 0;
}
/*}}}*/
